#!/usr/bin/perl 
################################################################################
#                                                                              #
#  2DHF version 1-2003                                                         #
#  Copyright (C) 2003  Jacek Kobus                                             #
#                                                                              #
#  This software may be used and distributed according to the terms            #
#  of the GNU General Public License, see README and COPYING.                  #
#                                                                              #
################################################################################
# Last update 10/11/03, 5/01/05 

# This script facilitates the calculation of polarizabilities and
# hyperpolarizabilities from a series of calculations performed by
# means of the 2D HF program.

# Usage:
 
#    ./elprop file_m2 file_m1 file_p0 file_p1 file_p2 

# where file_i are output files from x2dhf program corresponding to
# calculations with an external finite field -2F, -F, 0, +F and +2F,
# respectively.  F is extracted from the second of these files.

# The derivatives are evaluated by means of the 3 and 5-point central
# differences formulae in order to assess the accuracy of calculations.

# For further details see 

# J.~Kobus, D.~Moncrieff, and S.~Wilson, A comparison of the electric
# moments obtained from finite basis set and finite difference
# {Hartree-Fock} calculations for diatomic molecules, Phys. Rev. A 62
# (2000), 062503/1--9.

# J.~Kobus, D.~Moncrieff, and S.~Wilson, Comparison of the
# polarizabilities and hyperpolarizabilities obtained from finite
# basis set and finite field, finite difference {Hartree-Fock}
# calculations for diatomic molecules, J. Phys. B:
# At. Mol. Opt. Phys. 34 (2001) 5127-5143.

#######################################################################

# check the number of command line arguments

foreach ($i=0; $i<=4;$i++) {
    if ( $ARGV[$i] eq '' ) {
	print ("\nError: missing argument(s) \n\n");
	print ("Usage: ./elprop file_m2 file_m1 file_p0 file_p1 file_p2 \n\n"); 
	exit;
    }
}

$debug1=0; # set to nozero value for additional printouts
$debug2=1; # set to nozero value to monitor the dependence of alpha,
           # beta and gamma (h^4) values on the accuracy of the dipole moment


$narg=5;

$F=0;
$homo="no";
#$homo="no";
$ver=0;

# set to the absolute accuracy of the dipole moment

$autodelta="no";
$autodelta="yes";

$Delta=1e-12;   

# set to "yes" to get the summary of results in a LaTeX format
$latex="no";

# extract the strength of a field and some other parameters from the
# second of the files

$x2dhf_out=$ARGV[1]; 

if ($ARGV[5] == "") {
    open(X2DHF_OUT,"<$x2dhf_out") || die " cannot open: $!";
    while (<X2DHF_OUT>){
	if ( $_ =~ /finite electric field:/) {
	    @t=split(" ",$_);
	    $F=abs($t[3]);
	    last;
	} 
    }
    close(X2DHF_OUT);
} 
else {

# in case the F value is not included in an output file it can be
# given as the last argument

 	$F=$ARGV[5];
# 	print "zero","\n";
    }

# extract some other parameters from the file

open(X2DHF_OUT,"<$x2dhf_out") || die " cannot open: $!";
while (<X2DHF_OUT>){
    if ( $_ =~ /nuclei/i) {
# extract Z1 Z2 
	($t[1],$Z1,$Z2)=split(" ",$_);
    }

    if ( $_ =~ /molecular system:/) {
# extract the description of a case and the internuclear distance
	$_=<X2DHF_OUT>;
	$_=<X2DHF_OUT>;
	$header=$_;
	@t=split("=",$_);
	$t[2]=$t[1];
	@t=split(" ",$t[2]);
	$R=$t[0];
    }

    if ( $_ =~ /grid:/) {
# extract the grid size
	$_=<X2DHF_OUT>;
	$_ =~ /.*\s+(\d{2,})\s+\(/;
	$nu=$1;
	$_=<X2DHF_OUT>;
	$_ =~ /\s+(\d{2,})\s+\(/;
	$mu=$1;
	$_=<X2DHF_OUT>;
	$_ =~ /\s+(\d+\.\d\d)/;
	$R_inf=$1;
    }
    if ( $_ =~ /centre     atomic weight       z/ ) {
# extract atomic masses and the centre of mass
	$_=<X2DHF_OUT>;
	@t = split (" ",$_);
	$ma=$t[1];
	$_=<X2DHF_OUT>;
	@t = split (" ",$_);
	$mb=$t[1];
	$_=<X2DHF_OUT>;
	@t = split (" ",$_);
	$z=$t[1];
#	    if ($debug1) {print $ma,"  ",$mb, "  ",$z,"\n";}
    }

    if ( /electronic \(au\/Debye\-Ang\^k\)/ ) {	
	$ver=1; # output in the new format
    }
}
close(X2DHF_OUT);

# stop if the field strength has not been determined
if ($F == "" ) {
    print ("Error: finite field strength is zero \n");
    exit;
}

if ($ver == 0) {
    extract_data_0();
} else {
    extract_data_1();
}

sub extract_data_0 {

# examine the five output files (old format; versions < 1-2005) in
# turn

    for ($iarg=0; $iarg<$narg; $iarg++) {
	
	$x2dhf_out=$ARGV[$iarg]; 
	open(X2DHF_OUT,$x2dhf_out) || die " cannot open: $!";
	
	if ($debug1) {
	    printf "\n %6s %18s \n","File: ",$x2dhf_out;
	}
	
	while (<X2DHF_OUT>){
	    
	    if ( $_ =~ /centre     atomic weight       z/ ) {
# extract atomic masses and the centre of mass
		for ($i=1; $i<=7; $i++) {  # skip 7 lines 
		    $_=<X2DHF_OUT>;
		}
		
# extract dipole, quadrupole etc moments 
# electronic (au)           Q_1t    Q_2t  
# electronic (Debye-Ang**k) Q_1tda  Q_2tda 
# total (au)                Q_1     Q_2
# total (Debye-Ang**k)      Q_1da   Q_2da 
		
		if ($debug1) {
		    printf "%50s \n","multipole moments relative to centre of mass:";
		}
		
		@t = split (" ",$_);
		$Q_1t[$iarg]=$t[2]; $Q_2t[$iarg]=$t[3]; $Q_3t[$iarg]=$t[4]; $Q_4t[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","electronic (au)         : ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
		
		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1tda[$iarg]=$t[2]; $Q_2tda[$iarg]=$t[3]; $Q_3tda[$iarg]=$t[4]; $Q_4tda[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","electronic (Debye-Ang^k): ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
		
		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1[$iarg]=$t[2]; $Q_2[$iarg]=$t[3]; $Q_3[$iarg]=$t[4]; $Q_4[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","total (au)              : ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
		
		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1da[$iarg]=$t[2]; $Q_2da[$iarg]=$t[3]; $Q_3da[$iarg]=$t[4]; $Q_4da[$iarg]=$t[5];
		
		if ($debug1) {
		    printf "%33s","total (Debye-Ang^k)     : ";
		    for ($i=2; $i<=5; $i++) {
			printf "%15.6e", $t[$i];
		    }
		    printf "\n";
		}
	    }
	    
	    if ( $_ =~ / ... writing functions to disk .../) {
# extract total energy
		$_=<X2DHF_OUT>;
		$_=<X2DHF_OUT>;
		if ( $_ =~ /total energy/) {
		    ($t[0],$t[1],$t[2]) = split (" ",$_);
		    $t[2] =~ s/D/E/;
		    $E[$iarg]=$t[2];
		    if ($debug1) {
			printf "%18s %26.16e \n","total energy:",$E[$iarg];
		    }
		}
	    }
	}
	close(X2DHF_OUT);
    }
}


sub extract_data_1 {

# examine the five output files (new format, version >= 1-2005) in turn

    for ($iarg=0; $iarg<$narg; $iarg++) {
	
	$x2dhf_out=$ARGV[$iarg]; 
	open(X2DHF_OUT,$x2dhf_out) || die " cannot open: $!";
	
	if ($debug1) {
	    printf "\n %6s %18s \n","File: ",$x2dhf_out;
	}
	
	while (<X2DHF_OUT>){
	    
	    if ( $_ =~ /centre     atomic weight       z/ ) {
# extract atomic masses and the centre of mass
		for ($i=1; $i<=5; $i++) {  # skip 5 lines 
		    $_=<X2DHF_OUT>;
		}
		
# extract dipole, quadrupole etc moments 
# electronic (au)           Q_1t    Q_2t  
# electronic (Debye-Ang**k) Q_1tda  Q_2tda 
# total (au)                Q_1     Q_2
# total (Debye-Ang**k)      Q_1da   Q_2da 
		
		if ($debug1) {
		    printf "%50s \n","multipole moments relative to centre of mass:";
		}
		$_=<X2DHF_OUT>;
		@t = split (" ",$_);
		$Q_1t[$iarg]=$t[3]; $Q_1[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_1tda[$iarg]=$t[0]; $Q_1da[$iarg]=$t[1];

		$_=<X2DHF_OUT>; 		
		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_2t[$iarg]=$t[3]; $Q_2[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_2tda[$iarg]=$t[0]; $Q_2da[$iarg]=$t[1];

		$_=<X2DHF_OUT>; 		
		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_3t[$iarg]=$t[3]; $Q_3[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_3tda[$iarg]=$t[0]; $Q_3da[$iarg]=$t[1];


		$_=<X2DHF_OUT>; 		
		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_4t[$iarg]=$t[3]; $Q_4[$iarg]=$t[4];

		$_=<X2DHF_OUT>; 		
		@t = split (" ",$_);
		$Q_4tda[$iarg]=$t[0]; $Q_4da[$iarg]=$t[1];

		if ($debug1) {
		    printf "%42s",  "dipole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_1t[$iarg],$Q_1[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ',$Q_1tda[$iarg],$Q_1da[$iarg];


		    printf "\n%42s","quadrupole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_2t[$iarg],$Q_2[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ', $Q_2tda[$iarg],$Q_2da[$iarg];

		    printf "\n%42s","octopole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_2t[$iarg],$Q_2[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ', $Q_2tda[$iarg],$Q_2da[$iarg];

		    printf "\n%42s","hexadecapole electronic total (au/Debye): ";
		    printf "%15.6e%15.6e\n", $Q_2t[$iarg],$Q_2[$iarg];
		    printf "%42s%15.6e%15.6e\n", ' ', $Q_2tda[$iarg],$Q_2da[$iarg];
		}
	    }
	    
	    if ( $_ =~ / ... writing functions to disk .../) {
# extract total energy
		$_=<X2DHF_OUT>;
		$_=<X2DHF_OUT>;
		if ( $_ =~ /total energy/) {
		    ($t[0],$t[1],$t[2]) = split (" ",$_);
		    $t[2] =~ s/D/E/;
		    $E[$iarg]=$t[2];
		    if ($debug1) {
			printf "%18s %26.16e \n","total energy:",$E[$iarg];
		    }
		}
	    }
	}
	close(X2DHF_OUT);
    }
}

print "          ","\n";
print '=' x 73, "\n";
print " molecular system:          ","\n";
print ($header," \n");
print " grid:    [",$nu,"x",$mu,";",$R_inf,"]\n\n";

print " Z1        ",$Z1,"\n";
print " Z2        ",$Z2,"\n";
print " R         ",$R,"\n";
print " F         ",$F,"\n\n";

printf "%7s%+24.16f%16s%+22.16f\n", " E(-2F)",$E[0], "mu_z(-2F)",$Q_1[0];
printf "%7s%+24.16f%16s%+22.16f\n", " E(-F) ",$E[1], "mu_z(-F) ",$Q_1[1];
printf "%7s%+24.16f%16s%+22.16f\n", " E( 0) ",$E[2], "mu_z( 0) ",$Q_1[2];
printf "%7s%+24.16f%16s%+22.16f\n", " E(+F) ",$E[3], "mu_z(+F) ",$Q_1[3];
printf "%7s%+24.16f%16s%+22.16f\n", " E(+2F)",$E[4], "mu_z(+2F)",$Q_1[4];

print "\n Total multipole moments (F=0):\n";
#print '-' x 70,"\n";
$i=2;
printf "%14s%+22.16f\n","mu_z     ", $Q_1t[$i];
printf "%14s%+22.16f\n","Theta_zz ", $Q_2t[$i];
printf "%14s%+22.16f\n","Omega_zzz", $Q_3t[$i];
printf "%14s%+22.16f\n","Phi_zzzz ", $Q_4t[$i];

print "\n";
# z-component of the dipole moment calculated from the 3-point formula
$mu3=-($E[3]-$E[1])/(2.0*$F)+(-($R/2.0+$z)*$Z1+($R/2.0-$z)*$Z2);

# z-component of the dipole moment calculated from the 5-point formula
$mu5=($E[4]-8*$E[3]+8*$E[1]-$E[0])/(12.0*$F)+(-($R/2.0+$z)*$Z1+($R/2.0-$z)*$Z2);

$muq=$Q_1t[2];

printf "%5s  %19.16f %7s %22.16f %4s \n","mu_z",$mu3,"(E h^2)",$Q_1t[2],"(mu)    ";
printf "%5s  %19.16f %7s %22.16f %4s \n","mu_z",$mu5,"(E h^4)",$Q_1t[2],"(mu)    ";

if ($autodelta eq "yes") {
    $Delta = ($Q_1t[3]+$Q_1t[1]);
}

if ($debug2 != 0 ) {
    $change=$Q_1t[1]-$Delta;

    printf "\n";
    printf "%$20s %+20.16f \n", " mu_z(+F)        ",$Q_1t[3];
    printf "%$20s %+20.16f \n", " mu_z(-F)        ",$Q_1t[1];
    printf "%$20s %+20.1e  \n",  " Delta",$Delta;
    printf "%$20s %+20.16f \n", " mu_z(-F) + Delta",$change;
}

if ( $homo eq "yes" ) {
    $Q_1t[0] = -$Q_1t[4];
    $Q_1t[1] = -$Q_1t[3];
    $E[0]    =  $E[4];
    $E[1]    =  $E[3];
}

# alpha_zz calculated from the energy and dipole moment values by
# means of the 3-point formula (-1 0 1)

$alpha3e=-($E[3]+$E[1]-2.0*$E[2])/($F*$F);
$alpha3q= ($Q_1t[3]-$Q_1t[1])/(2.0*$F);

# alpha_zz calculated from the energy and dipole moment values by
# means of the 5-point formula (1 -8  8 -1)

$alpha5e= ($E[4]-16*$E[3]+30*$E[2]-16*$E[1]+$E[0])/(12*$F*$F);
$alpha5q= (-$Q_1t[4]+8*$Q_1t[3]-8*$Q_1t[1]+$Q_1t[0])/(12.0*$F);

print "          ","\n";
printf "%9s  %23.16f %7s %20.16f %6s \n","alpha_zz",$alpha3e,"(E h^2)",$alpha3q,"(mu h^2)";
printf "%9s  %23.16f %7s %20.16f %6s \n","alpha_zz",$alpha5e,"(E h^4)",$alpha5q,"(mu h^4)";

if ($debug2 != 0 ) {
    $change=$Q_1t[1]+$Delta;                
    $dalpha5q= (-$Q_1t[4]+8*$Q_1t[3]-8*$change+$Q_1t[0])/(12.0*$F);
    printf "%16s %46.16f %6s \n", "alpha_zz(Delta)",$dalpha5q,"(mu h^4)";
}

# zzz-component of the first hyperpolarizability calculated from the
# dipole moment values by means of the 3 and 5-point formulae (1 -2  1
# and -1 16 -30 16 -1)

$beta5e=(-$E[4]+2*$E[3]-2*$E[1]+$E[0])/(2*$F*$F*$F);
$beta3q=  ($Q_1t[3]+$Q_1t[1]-2.0*$Q_1t[2])/($F*$F);
$beta5q=  (-$Q_1t[4]+16*$Q_1t[3]-30*$Q_1t[2]+16*$Q_1t[1]-$Q_1t[0])/(12*$F*$F);

print ("\n");
printf "%9s  %52.16f %6s \n","beta_zzz",$beta3q,"(mu h^2)";
printf "%9s  %23.16f %7s %20.16f %6s \n","beta_zzz",$beta5e,"(E h^4)",$beta5q,"(mu h^4)";
#printf "%9s  %52.16f %6s \n","beta_zzz",$beta5q,"(mu h^4)";

if ($debug2 != 0 ) {
    $change=$Q_1t[1]+$Delta;                
    $dbeta5q=-($Q_1t[4]-16*$Q_1t[3]+30*$Q_1t[2]-16*$change+$Q_1t[0])/(12*$F*$F);
    printf "%16s %46.16f %6s \n", "beta_zzz(Delta)",$dbeta5q,"(mu h^4)";
}

$debug3=0;
if ($debug3 != 0) {
    print ("\n");
    print ($Q_1[2],"  ",$Q_1t[2], "\n");
    print ($Q_1t[4], "   ", -$Q_1t[4]+$Q_1t[0],"\n");
    
    print ($Q_1t[3], "   ",$Q_1t[3]-$Q_1t[1]," \n");
    
    print ((-$Q_1t[4]+$Q_1t[0]), "  ",2*($Q_1t[3]-$Q_1t[1])," \n");
    
    print (((-$Q_1t[4]+$Q_1t[0])+2*($Q_1t[3]-$Q_1t[1]))/2," \n");
}

# zzzz-component of the second hyperpolarizability calculated from the
# dipole moment values by means of the 5-point formulae

$gamma5q=-(-$Q_1t[4]+2*$Q_1t[3]-2*$Q_1t[1]+$Q_1t[0])/(2*$F*$F*$F);

printf "\n";
printf "%11s  %50.16f %6s \n","gamma_zzzz",$gamma5q,"(mu h^4)";

if ($debug2 != 0 ) {
    $change=$Q_1t[1]+$Delta;                
    $dgamma5q=-(-$Q_1t[4]+2*$Q_1t[3]-2*$change+$Q_1t[0])/(2*$F*$F*$F);
    printf "%18s %44.16f %6s \n", "gamma_zzzz(Delta)",$dgamma5q,"(mu h^4)";
}
print '=' x 73, "\n";

print "Summary:\n";
$width=14;
$acc1=10;
$acc2=5;
$acc3=4;
$acc4=1;

print $header,"\n";
#$header =~ /\s+(\w\w?)\s*(\w?\w?)\s+\(/;
#$header =~ /\s+([A-Z][a-z]?)\s*(\w?\w?)\s+\(/;
#$system = $1 . $2;
$header =~ s/R = .+//;
$header =~ s/[\d\W]//g;

printf "%28s\n",$header;
printf "%18s %+$width.${acc1}f\n", "mu_z      ",$muq;
printf "%18s %+$width.${acc2}f\n", "alpha_zz  ",$alpha5q;
printf "%18s %+$width.${acc3}f\n", "beta_zzz  ",$beta5q;
printf "%18s %+$width.${acc4}f\n", "gamma_zzzz",$gamma5q;



if ( $latex eq "yes" ) {
    print "\n";
    printf "%26s %3s\n","&",$header;
    printf "%18s &%+$width.${acc1}f\n", "\\mu_{z}      ",$muq;
    printf "%18s &%+$width.${acc2}f\n", "\\alpha_{zz}  ",$alpha5q;
    printf "%18s &%+$width.${acc3}f\n", "\\beta_{zzz}  ",$beta5q;
    printf "%18s &%+$width.${acc4}f\n", "\\gamma_{zzzz}",$gamma5q;
}
$debug=0;

if ( $debug != 0 ) {

    print " Total multipole moments:\n";
    for ($i=0; $i<=4; $i++) {
	print '-' x 70,"\n";
#	print " File ",$i+1,"\n";
	print " Q_1  ",  $Q_1t[$i], " au     ", $Q_1da[$i],  " Debye-Ang^0 \n";
	print " Q_2  ",  $Q_2t[$i], " au     ", $Q_2da[$i],  " Debye-Ang^1 \n";
	print " Q_3  ",  $Q_3t[$i], " au     ", $Q_3da[$i],  " Debye-Ang^2 \n";
	print " Q_4  ",  $Q_4t[$i], " au     ", $Q_4da[$i],  " Debye-Ang^3 \n";
    }

    print '-' x 70,"\n";
    print "\n Electronic multipole moments:\n";
    for ($i=0; $i<=4; $i++) {
	print '-' x 70,"\n";
	print " File ",$i+1,"\n";
	print " Q_1t  ",  $Q_1[$i], " au     ", $Q_1tda[$i],  " Debye-Ang^0 \n";
	print " Q_2t  ",  $Q_2[$i], " au     ", $Q_2tda[$i],  " Debye-Ang^1 \n";
	print " Q_3t  ",  $Q_3[$i], " au     ", $Q_3tda[$i],  " Debye-Ang^2 \n";
	print " Q_4t  ",  $Q_4[$i], " au     ", $Q_4tda[$i],  " Debye-Ang^3 \n";
    }
	print '-' x 70,"\n";
}
